/*
  "DataTransmitter"
  Library supporting wireless or wire-bound serial data broadcasting
  (e.g. sensor data via a 2-wire connection or radio broadcast on an ISM-frequency)
  The transmission is DC-free (Manchester-coded) and error-checked (CRC8)
  Requires library "CRCGenerator"
  
  V2.1 / Copyright (C) 2023, T.Rode (DL1DUZ)

  Permission is granted to use, copy, modify, and distribute this software
  and documentation for non-commercial purposes.

  Required ressources:
   - 1 digital output pin
   - Timer 2 (conflicting with the functions  "analogWrite()" at pins 3 & 11 and "tone()" (all pins)).
*/

//################################################################################################################
//includes
//################################################################################################################

#include <DataTransmitter.h>

//################################################################################################################
//declarations
//################################################################################################################

//################################################################################################################
//global functions
//##############################################################################################################

// compare-isr (called by timer2)
ISR(TIMER2_COMPA_vect) {
  DT.isr();
}

//################################################################################################################
//private functions
//##############################################################################################################

//################################################################################################################
//public functions
//################################################################################################################

// performs Manchester-coded transmission of dataset (LSB first)
void DataTransmitterClass::isr() {

  // each symbol (bit) requires two steps to transmit (1 will be transmitted as 0|1 and 0 as 1|0
  if(step & 1) { // is odd step number -> first step out of two
    current_symbol = current_byte & 1; // load symbol (LSB)
    current_byte >>= 1; // move on to next symbol
    // toggle output if symbol-value (bit-value) has changed
    if(prev_symbol ^ current_symbol) {
      pin_value = !pin_value;
    }
    prev_symbol = current_symbol;
  }

  if(step == block_size) { // end of data block (header or byte of user data) reached
    step = 0;
    if(byte_counter == bytes) { // end of user data reached
      if(block_size == 2) {
        // terminate transmission by stopping timer 2
        TCCR2B = 0; // set entire TCCR2B register to 0
        bytes = 0;
      }
      else { // prepare for termination
        current_byte = 0; // transmit 0 as last symbol to force 1|0 at end of transmission
        block_size = 2; // equals 1 bit
      }
    }
    else { // load next byte of user data
      current_byte = *(data+byte_counter);
      block_size = 16; // equals one byte of user data
      ++byte_counter;
    }
  }

  ++step;
  pin_value = !pin_value;

  digitalWrite(pin, pin_value ^ is_low_active);

}

//################################################################################################################

/*
initializes the library

"BYTE_RATE" -> the target gross data transfer rate in Byte/s (the algorithm will attempt getting as close as possible
 to this number). Valid entries are 4...255. A small overhead is introduced by header data (2 Byte for CRC & ID
 + 5 bit for synchronisation)
"TRANSMITTER_PIN" -> the digital pin the transmitter is connected to
"IS_LOW_ACTIVE" -> if True, the transmitter-logic is low-active (low->TX on)
*/
void DataTransmitterClass::init(const uint8_t BYTE_RATE, const uint8_t TRANSMITTER_PIN, const uint8_t IS_LOW_ACTIVE) {

  pin = TRANSMITTER_PIN;
  is_low_active = IS_LOW_ACTIVE;

  pinMode(pin, OUTPUT);
  digitalWrite(pin, is_low_active);

  // ... the Timer 2 setup ...

  // calculate time multiplier at 16MHz clock speed from given Byte-rate (e.g. 50 Byte/s equals 400 bit/s or 800 pulses/s)
  const uint8_t PRE[7] = {0, 3, 2, 1, 1, 1, 2}; // number of bits to right-shift multiplier at each loop
  uint8_t i = 0;
  multiplier = 1000000/max(BYTE_RATE, 4); // e.g. 250000, 20000 and 3921 at 4, 50 and 255 Byte/s
  // determine prescaler and calculate OCR2A-value
  do {
    multiplier>>=PRE[i];
    ++i;
  }
  while(multiplier > 0x100); // until multiplier fits into 8 bit
  prescaler = i; //prescaling equals TCCR2B(0,0,0,0,0,x,x,x) with 1=system clock, 2=8, 3=32, 4=64, 5=128, 6=256, 7=1024
                // e.g 1024, 128 and 32 at 4, 50 and 255 Bytes/s
  --multiplier; // e.g. 243, 155 and 121 at 4, 50 and 255 Bytes/s
}

//################################################################################################################

/*
creates a structure of the type dset_t (transmitter dataset parameter)

"X" -> a pointer to the data-element (can be a single variable, array, ...)
"ID" -> the ID of the dataset (0...255)
"SCOPE" -> the number of data elements times their size (1...255 Byte)
*/

dset_t DataTransmitterClass::createDataset(const void* X, const uint8_t ID, const uint8_t SCOPE) {
  dset_t dummy;

  dummy.data = (uint8_t*) X;
  dummy.ID = ID;
  dummy.SIZE = SCOPE;

  return dummy;
}

//################################################################################################################

/*
transmits a set of Byte-data including ID and calculated CRC8-value for validation (LSB first)
in case of an ongoing transmission or SIZE == 0 this function will return 0, otherwise 1

"DSET" -> a pointer to the dataset
*/

uint8_t DataTransmitterClass::transmitData(const dset_t DSET) {
  uint8_t commenced = 0;

  if(!bytes && DSET.SIZE) {
    bytes = DSET.SIZE;
    data = DSET.data;
    // load current byte with header (five sync-bits + ID + CRC)
    current_byte = CRC.crcCalculation(DSET.ID, DSET.data, DSET.SIZE);
    current_byte <<= 8;
    current_byte += DSET.ID;
    current_byte <<= 5;
    current_byte += 0b00010000; // add the five sync-bits at LSB

    step = 1; // each symbol requires two steps to transmit
    prev_symbol = 0; // equal to first symbol (bit) to be transmitted (which is 0)
    byte_counter = 0;
    block_size = 42; // marks end of header-block (2*(16bit + 5bit))
    pin_value = 0;
  
    // init & start timer 2
    cli(); // disable global interrupts
    TCCR2A = 0; // set entire TCCR2A register to 0
    TCCR2B = 0; // set entire TCCR2B register to 0
    TCNT2  = 0; //initialize counter value to 0
    OCR2A = multiplier; // set output compare register depending on the calculated trigger time
    TCCR2B = prescaler; // set prescaler
    TCCR2A |= (1 << WGM21); // enable timer CTC mode
    TIMSK2 |= (1 << OCIE2A); // enable timer compare interrupt
    sei(); // enable global interrupts

    commenced = 1;
  }

  return commenced;
}

//################################################################################################################

/*
reads/writes an arbitrary type of data from/to the dataset
this function will return 0 in case of an error or 255 if the array goes out of scope, otherwise it will return
the next position in the data array

"X" -> a pointer to the data-element to be read/written
"SCOPE" -> the number of data elements times their size
"DSET" -> a pointer to the dataset
"POS" -> the position from which on the data element should be stored or read
"WRITE" -> flag indicating that date should be written into the array
*/

uint8_t DataTransmitterClass::dataTransfer(const void* X, const uint16_t SCOPE, const dset_t DSET, const uint8_t POS,
                                           const uint8_t WRITE) {

  uint8_t next = 0;

  if(SCOPE && (SCOPE <= DSET.SIZE-POS)) {
    for(uint16_t i=0; i<SCOPE; ++i) {
      if(WRITE) {
        *(DSET.data + POS + i) = *(((uint8_t*) X) + i);
      }
      else {
        *(((uint8_t*) X) + i) = *(DSET.data + POS + i);
      }
    }
    next = POS + SCOPE;
    if(next == DSET.SIZE) { next = 0xFF; }
  }

  return next;
}

DataTransmitterClass DT;
